@@ -16,6 +16,8 @@ import atexit
 import builtins
 import cloudpickle
 import dis
+import functools
+import inspect
 import os
 import pathlib
 import pickle
@@ -31,7 +33,7 @@ import threading
 import time
 import traceback
 from collections import defaultdict
-from functools import lru_cache
+from functools import lru_cache, wraps
 from rich.console import Console
 from rich.markdown import Markdown
 from rich.syntax import Syntax
@@ -59,17 +61,9 @@ from scalene.adaptive import Adaptive
 from scalene.runningstats import RunningStats
 from scalene import sparkline
 
-# Logic to ignore @profile decorators.
-try:
-    builtins.profile  # type: ignore
-except AttributeError:
-
-    def profile(func: Any) -> Any:
-        """No line profiler; we provide a pass-through version."""
-        return func
-
-    builtins.profile = profile  # type: ignore
-
+Filename = NewType("Filename", str)
+LineNumber = NewType("LineNumber", int)
+ByteCodeIndex = NewType("ByteCodeIndex", int)
 
 assert (
     sys.version_info[0] == 3 and sys.version_info[1] >= 6
@@ -204,7 +198,8 @@ def parse_args() -> Tuple[argparse.Namespace, List[str]]:
 
 arguments, left = parse_args()
 
-# Load shared objects unless the user specifies "--cpu-only" at the command-line.
+# Load shared objects (that is, interpose on malloc, memcpy and friends)
+# unless the user specifies "--cpu-only" at the command-line.
 # (x86-64 only for now.)
 
 if not arguments.cpu_only and (
@@ -231,7 +226,10 @@ if (
             args = [os.path.basename(sys.executable), "-m", "scalene"] + args
             result = subprocess.run(args)
             if result.returncode < 0:
-                print("scalene error: received signal", signal.Signals(-result.returncode).name)
+                print(
+                    "scalene error: received signal",
+                    signal.Signals(-result.returncode).name,
+                )
             sys.exit(result.returncode)
 
     # Similar logic, but for Mac OS X.
@@ -247,13 +245,18 @@ if (
             args = [os.path.basename(sys.executable), "-m", "scalene"] + args
             result = subprocess.run(args, close_fds=True, shell=False)
             if result.returncode < 0:
-                print("scalene error: received signal", signal.Signals(-result.returncode).name)
+                print(
+                    "scalene error: received signal",
+                    signal.Signals(-result.returncode).name,
+                )
             sys.exit(result.returncode)
 
-Filename = NewType("Filename", str)
-LineNumber = NewType("LineNumber", int)
-ByteCodeIndex = NewType("ByteCodeIndex", int)
+# Install our profile decorator.
 
+def scalene_redirect_profile(func):
+    return Scalene.profile(func)
+
+builtins.profile = scalene_redirect_profile
 
 class Scalene:
     """The Scalene profiler itself."""
@@ -261,6 +264,12 @@ class Scalene:
     # Debugging flag, for internal use only.
     __debug: bool = False
 
+    # Support for @profile
+    # decorated files
+    __files_to_profile: Dict[Filename, bool] = defaultdict(bool)
+    # decorated functions
+    __functions_to_profile: Dict[Filename, Dict[Any, bool]] = defaultdict(lambda: {})
+
     # We use these in is_call_function to determine whether a
     # particular bytecode is a function call.  We use this to
     # distinguish between Python and native code execution when
@@ -278,6 +287,7 @@ class Scalene:
         [threading.Thread, Union[builtins.float, None]], None
     ] = threading.Thread.join
 
+    # As above; we'll cache the original thread and replace it.
     __original_lock = threading.Lock
 
     @staticmethod
@@ -407,9 +417,21 @@ class Scalene:
     # Things that need to be in sync with include/sampleheap.hpp:
     #
     #   file to communicate the number of malloc/free samples (+ PID)
-    __malloc_signal_filename = Filename("/tmp/scalene-malloc-signal")
+    __malloc_signal_filename = Filename("/tmp/scalene-malloc-signal" + str(os.getpid()))
+    __malloc_signal_position = 0
+    try:
+        __malloc_signal_fd = open(__malloc_signal_filename, "x")
+    except BaseException:
+        pass
+    __malloc_signal_fd = open(__malloc_signal_filename, "r")
+    
     #   file to communicate the number of memcpy samples (+ PID)
-    __memcpy_signal_filename = Filename("/tmp/scalene-memcpy-signal")
+    __memcpy_signal_filename = Filename("/tmp/scalene-memcpy-signal" + str(os.getpid()))
+    try:
+        __memcpy_signal_fd = open(__memcpy_signal_filename, "r")
+        __memcpy_signal_position = 0
+    except BaseException:
+        pass
 
     # The specific signals we use.
     # Malloc and free signals are generated by include/sampleheap.hpp.
@@ -439,7 +461,7 @@ class Scalene:
     #   the name of the program being profiled
     __program_being_profiled = Filename("")
 
-    # Is the thread sleeping? (We use this in to properly attribute CPU time.)
+    # Is the thread sleeping? (We use this to properly attribute CPU time.)
     __is_thread_sleeping: Dict[int, bool] = defaultdict(bool)  # False by default
 
     # Threshold for highlighting lines of code in red.
@@ -451,6 +473,22 @@ class Scalene:
     # Default threshold for number of mallocs to report a file.
     __malloc_threshold = 100
 
+    # Replacement @profile decorator function.
+    # We track which functions - in which files - have been decorated,
+    # and only report stats for those.
+    @staticmethod
+    def profile(func: Any) -> Any:
+        # Record the file and function name
+        Scalene.__files_to_profile[func.__code__.co_filename] = True
+        Scalene.__functions_to_profile[func.__code__.co_filename][func] = True
+
+        @functools.wraps(func)
+        def wrapper_profile(*args: Any, **kwargs: Any) -> Any:
+            value = func(*args, **kwargs)
+            return value
+
+        return wrapper_profile
+
     @staticmethod
     def set_thread_sleeping(tid: int) -> None:
         Scalene.__is_thread_sleeping[tid] = True
@@ -550,6 +588,7 @@ start the timer interrupts."""
         """Replace lock with a version that periodically yields and updates sleeping status."""
 
         def __init__(self) -> None:
+            # Cache the original lock (which we replace)
             self.__lock: threading.Lock = Scalene.get_original_lock()
 
         def acquire(self, blocking: bool = True, timeout: float = -1) -> bool:
@@ -595,13 +634,6 @@ start the timer interrupts."""
         threading.Thread.join = Scalene.thread_join_replacement  # type: ignore
         # Hijack lock.
         threading.Lock = Scalene.ReplacementLock  # type: ignore
-        # Build up signal filenames (adding PID to each).
-        Scalene.__malloc_signal_filename = Filename(
-            Scalene.__malloc_signal_filename + str(os.getpid())
-        )
-        Scalene.__memcpy_signal_filename = Filename(
-            Scalene.__memcpy_signal_filename + str(os.getpid())
-        )
         if "cpu_percent_threshold" in arguments:
             Scalene.__cpu_percent_threshold = int(arguments.cpu_percent_threshold)
         if "malloc_threshold" in arguments:
@@ -688,6 +720,21 @@ start the timer interrupts."""
             Scalene.cpu_signal_handler_helper(signum, this_frame)
             Scalene.__in_signal_handler.release()
 
+    @staticmethod
+    def profile_this_code(fname: Filename, lineno: LineNumber) -> bool:
+        """When using @profile, only profile files & lines that have been decorated."""
+        if not Scalene.__files_to_profile:
+            return True
+        if fname not in Scalene.__files_to_profile:
+            return False
+        # Now check to see if it's the right line range.
+        for fn in Scalene.__functions_to_profile[fname]:
+            lines, line_start = inspect.getsourcelines(fn)
+            if lineno >= line_start and lineno < line_start + len(lines):
+                # Yes, it's in range.
+                return True
+        return False
+
     @staticmethod
     def cpu_signal_handler_helper(
         _signum: Union[Callable[[Signals, FrameType], None], int, Handlers, None],
@@ -893,30 +940,28 @@ start the timer interrupts."""
         if not new_frames:
             return
 
-        # Process the input array.
+        # Process the input array from where we left off reading last time.
         arr: List[Tuple[int, str, float, float]] = []
         try:
-            with open(Scalene.__malloc_signal_filename, "r") as mfile:
-                for count_str in mfile:
-                    count_str = count_str.rstrip()
+            mfile = Scalene.__malloc_signal_fd
+            mfile.seek(Scalene.__malloc_signal_position)
+            for count_str in mfile:
+                count_str = count_str.rstrip()
+                (
+                    action,
+                    alloc_time_str,
+                    count_str,
+                    python_fraction_str,
+                ) = count_str.split(",")
+                arr.append(
                     (
+                        int(alloc_time_str),
                         action,
-                        alloc_time_str,
-                        count_str,
-                        python_fraction_str,
-                    ) = count_str.split(",")
-                    arr.append(
-                        (
-                            int(alloc_time_str),
-                            action,
-                            float(count_str),
-                            float(python_fraction_str),
-                        )
+                        float(count_str),
+                        float(python_fraction_str),
                     )
-        except FileNotFoundError:
-            pass
-        try:
-            os.remove(Scalene.__malloc_signal_filename)
+                )
+            Scalene.__malloc_signal_position = mfile.tell()
         except FileNotFoundError:
             pass
 
@@ -944,10 +989,10 @@ start the timer interrupts."""
 
         for (frame, _tident, _orig_frame) in new_frames:
             fname = Filename(frame.f_code.co_filename)
-            line_no = LineNumber(frame.f_lineno)
+            lineno = LineNumber(frame.f_lineno)
             bytei = ByteCodeIndex(frame.f_lasti)
             # Add the byte index to the set for this line (if it's not there already).
-            Scalene.__bytei_map[fname][line_no].add(bytei)
+            Scalene.__bytei_map[fname][lineno].add(bytei)
             curr = before
             python_frac = 0.0
             allocs = 0.0
@@ -962,22 +1007,22 @@ start the timer interrupts."""
                     python_frac += python_fraction * count
                 else:
                     curr -= count
-                Scalene.__per_line_footprint_samples[fname][line_no].add(curr)
+                Scalene.__per_line_footprint_samples[fname][lineno].add(curr)
             assert curr == after
             # If there was a net increase in memory, treat it as if it
             # was a malloc; otherwise, treat it as if it was a
             # free. This is for later reporting of net memory gain /
             # loss per line of code.
             if after > before:
-                Scalene.__memory_malloc_samples[fname][line_no][bytei] += after - before
-                Scalene.__memory_python_samples[fname][line_no][bytei] += (
+                Scalene.__memory_malloc_samples[fname][lineno][bytei] += after - before
+                Scalene.__memory_python_samples[fname][lineno][bytei] += (
                     python_frac / allocs
                 ) * (after - before)
                 Scalene.__malloc_samples[fname] += 1
                 Scalene.__total_memory_malloc_samples += after - before
             else:
-                Scalene.__memory_free_samples[fname][line_no][bytei] += before - after
-                Scalene.__memory_free_count[fname][line_no][bytei] += 1
+                Scalene.__memory_free_samples[fname][lineno][bytei] += before - after
+                Scalene.__memory_free_count[fname][lineno][bytei] += 1
                 Scalene.__total_memory_free_samples += before - after
 
     @staticmethod
@@ -997,15 +1042,13 @@ start the timer interrupts."""
         # Process the input array.
         arr: List[Tuple[int, int]] = []
         try:
-            with open(Scalene.__memcpy_signal_filename, "r") as mfile:
-                for count_str in mfile:
-                    count_str = count_str.rstrip()
-                    (memcpy_time_str, count_str2) = count_str.split(",")
-                    arr.append((int(memcpy_time_str), int(count_str2)))
-        except FileNotFoundError:
-            pass
-        try:
-            os.remove(Scalene.__memcpy_signal_filename)
+            mfile = Scalene.__memcpy_signal_fd
+            mfile.seek(Scalene.__memcpy_signal_position)
+            for count_str in mfile:
+                count_str = count_str.rstrip()
+                (memcpy_time_str, count_str2) = count_str.split(",")
+                arr.append((int(memcpy_time_str), int(count_str2)))
+            Scalene.__memcpy_signal_position = mfile.tell()
         except FileNotFoundError:
             pass
         arr.sort()
@@ -1026,6 +1069,11 @@ start the timer interrupts."""
     @lru_cache(None)
     def should_trace(filename: str) -> bool:
         """Return true if the filename is one we should trace."""
+        # If the @profile decorator has been used,
+        # we restrict profiling to files containing decorated functions.
+        if Scalene.__files_to_profile:
+            return filename in Scalene.__files_to_profile
+        # Generic handling follows (when no @profile decorator has been used).
         if not filename:
             return False
         if filename[0] == "<":
@@ -1062,6 +1110,8 @@ start the timer interrupts."""
         fname: Filename, line_no: LineNumber, line: str, console: Console, tbl: Table,
     ) -> bool:
         """Print at most one line of the profile (true == printed one)."""
+        if not Scalene.profile_this_code(fname, line_no):
+            return False
         current_max = Scalene.__max_footprint
         did_sample_memory: bool = (
             Scalene.__total_memory_free_samples + Scalene.__total_memory_malloc_samples
